

from constants import UNKNOWN, COMPUTE, MEMORY, STORAGE, NETWORK

from base      import AtomBase
from compute   import Compute
from memory    import Memory
from storage   import Storage
from network   import Network

from _atoms    import atom_compute_asm 
from _atoms    import atom_compute
from _atoms    import atom_memory
from _atoms    import atom_storage 
from _atoms    import atom_network 





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils.signatures   as rus

from   _atoms    import atom_compute_asm
from   _atoms    import atom_compute
from   base      import AtomBase
from   constants import COMPUTE

OVERHEAD = 20   # in %

# ------------------------------------------------------------------------------
#
class Compute (AtomBase) :
    """
    This Compute Synapse emulates a compute workload, i.e. it consumes
    a specified number of floating point operations.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Compute')
    @rus.returns (rus.nothing)
    def __init__ (self) : 

        AtomBase.__init__ (self, COMPUTE)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Compute', list)
    @rus.returns (rus.nothing)
    def emulate (self, vals) : 

        print 'cpu: %s' % vals

        ops = int(vals[0] * 0.6)

        # remove a empirical overhead
      # ops -= int(ops/100*OVERHEAD)
        self._run (ops)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Compute', int)
    @rus.returns (rus.nothing)
    def _emulate (self, ops) : 

        atom_compute_asm (ops)


#-------------------------------------------------------------------------------






__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import os
import time
import errno
import psutil
import threading
import subprocess
import multiprocessing

import radical.utils              as ru
import radical.utils.logger       as rul
import radical.utils.signatures   as rus

from constants import UNKNOWN, COMPUTE, STORAGE, NETWORK


# ------------------------------------------------------------------------------
#
class AtomBase (object) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase', 
                  basestring, 
                  dict)
    @rus.returns (rus.nothing)
    def __init__  (self, atype) :


        self._atype = atype
        self._pid   = os.getpid ()
        self._uid   = ru.generate_id ("%s" % self._atype)
        self.logger = rul.getLogger ("radical.synapse.%s" % self._uid)


        # storage for temporary data and statistics
     #  self._tmpdir = "/scratch/synapse/" # FIXME
        self._tmpdir = "/tmp/"             # FIXME

        try:
            os.makedirs (self._tmpdir)
        except OSError as exc :
            if exc.errno == errno.EEXIST and os.path.isdir (self._tmpdir) :
                pass
            else: raise


        # start worker process
        self._work_queue   = multiprocessing.Queue ()
        self._result_queue = multiprocessing.Queue ()

        self._proc  = threading.Thread (target=self.run)
        self._proc.start ()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase')
    @rus.returns (basestring)
    def __str__  (self) :

        return self._uid


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase')
    @rus.returns (rus.nothing)
    def run (self) :

        try:

            while True :

                data = self._work_queue.get ()

                if data == None:
                    # signal to finish
                    return

                print "emulate %-20s: %s" % (self, str(data))
                self._emulate (*data)
                self._result_queue.put (True)


        except Exception as e:
            print "main loop error in atom driver: %s" % e
            raise


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase')
    @rus.returns (rus.nothing)
    def _run (self, *args) : 

        self._work_queue.put (args)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase')
    @rus.returns (dict)
    def wait (self) :

        return self._result_queue.get ()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('AtomBase')
    @rus.returns (rus.nothing)
    def stop (self) :

        self._work_queue.put (None) # signal finish

        if  self._proc :
            self._proc.join ()


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils.signatures   as rus

from   _atoms    import atom_storage
from   base      import AtomBase
from   constants import STORAGE

# ------------------------------------------------------------------------------
#
class Storage (AtomBase) :
    """
    This Storage Synapse emulates a storage workload, i.e. it allocates
    a specified storage size on disk.  It creates only minimal I/O to the
    allocated storage, to ensure that the full amount has been allocated.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Storage')
    @rus.returns (rus.nothing)
    def __init__ (self) : 

        AtomBase.__init__ (self, STORAGE)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Storage', list)
    @rus.returns (rus.nothing)
    def emulate (self, vals) : 

        rsize = int(vals[0])
        wsize = int(vals[1])

        src = "/tmp/synapse_storage.in"
        tgt = "/tmp/synapse_storage.%(pid)s.out"

        src = src % { 'tmp' : self._tmpdir, 'pid' : self._pid   }
        tgt = tgt % { 'tmp' : self._tmpdir, 'pid' : self._pid   }

        self._run (src, rsize, tgt, wsize)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Storage', basestring, int, basestring, int)
    @rus.returns (rus.nothing)
    def _emulate (self, src, rsize, tgt, wsize) : 

        atom_storage (src, rsize, tgt, wsize)


#-------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils.signatures   as rus

from   _atoms    import atom_memory
from   base      import AtomBase
from   constants import MEMORY

# ------------------------------------------------------------------------------
#
class Memory (AtomBase) :
    """
    This Memory Synapse emulates a memory workload, i.e. it allocates
    a specified memory size.  It does not create nor consume any I/O to the
    allocated memory -- instead, it quits right after allocation.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Memory')
    @rus.returns (rus.nothing)
    def __init__ (self) : 

        AtomBase.__init__ (self, MEMORY)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Memory', list)
    @rus.returns (rus.nothing)
    def emulate (self, vals) : 

        size = int(vals[0])

        return self._run (size)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Memory', int)
    @rus.returns (rus.nothing)
    def _emulate (self, size) : 

        try:
         #  print "atom_memory (%s)" % size
            atom_memory (size)
        except Exception as e:
            print "mem atom error: %s" % e



#-------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"



# ------------------------------------------------------------------------------

from synapsify import synapsify
from synapsify import NOTHING
from synapsify import PROFILE
from synapsify import EMULATE

from synapse   import profile
from synapse   import emulate
from synapse   import execute



# ------------------------------------------------------------------------------
#

import os

import radical.utils        as ru
import radical.utils.logger as rul


pwd     = os.path.dirname (__file__)
root    = "%s/.." % pwd
version, version_detail, version_branch, sdist_name, sdist_path = ru.get_version ([root, pwd])

_logger = rul.logger.getLogger  ('radical.synapse')
_logger.info ('radical.synapse      version: %s' % version_detail)


# ------------------------------------------------------------------------------





__author__    = 'RADICAL Team'
__email__     = 'radical@rutgers.edu'
__copyright__ = 'Copyright 2013/14, RADICAL Research, Rutgers University'
__license__   = 'MIT'


""" Setup script. Used by easy_install and pip. """

import os
import sys
import glob
import subprocess as sp

from setuptools           import setup, Command, find_packages
from setuptools.extension import Extension

name     = 'radical.synapse'
mod_root = 'src/radical/synapse'

# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version:          1.2.3            - is used for installation
#   - version_detail:  v1.2.3-9-g0684b06 - is used for debugging
#   - version is read from VERSION file in src_root, which then is copied to
#     module dir, and is getting installed from there.
#   - version_detail is derived from the git tag, and only available when
#     installed from git.  That is stored in mod_root/VERSION in the install
#     tree.
#   - The VERSION file is used to provide the runtime version information.
#
def get_version (mod_root):
    """
    mod_root
        a VERSION file containes the version strings is created in mod_root,
        during installation.  That file is used at runtime to get the version
        information.  
        """

    try:

        version        = None
        version_detail = None

        # get version from './VERSION'
        src_root = os.path.dirname (__file__)
        if  not src_root :
            src_root = '.'

        with open (src_root + '/VERSION', 'r') as f :
            version = f.readline ().strip()


        # attempt to get version detail information from git
        p   = sp.Popen ('cd %s ; '\
                        'tag=`git describe --tags --always` 2>/dev/null ; '\
                        'branch=`git branch | grep -e "^*" | cut -f 2 -d " "` 2>/dev/null ; '\
                        'echo $tag@$branch'  % src_root,
                        stdout=sp.PIPE, stderr=sp.STDOUT, shell=True)
        version_detail = p.communicate()[0].strip()

        if  p.returncode   !=  0  or \
            version_detail == '@' or \
            'fatal'        in version_detail :
            version_detail =  'v%s' % version

        print 'version: %s (%s)' % (version, version_detail)


        # make sure the version files exist for the runtime version inspection
        path = '%s/%s' % (src_root, mod_root)
        print 'creating %s/VERSION' % path
        with open (path + "/VERSION", "w") as f : f.write (version_detail + "\n")

        sdist_name = "%s-%s.tar.gz" % (name, version_detail)
        sdist_name = sdist_name.replace ('/', '-')
        sdist_name = sdist_name.replace ('@', '-')
        sdist_name = sdist_name.replace ('#', '-')
        if '--record'  in sys.argv or 'bdist_egg' in sys.argv :   
           # pip install stage 2      easy_install stage 1
           # NOTE: pip install will untar the sdist in a tmp tree.  In that tmp
           # tree, we won't be able to derive git version tags -- so we pack the
           # formerly derived version as ./VERSION
            os.system ("mv VERSION VERSION.bak")        # backup version
            os.system ("cp %s/VERSION VERSION" % path)  # use full version instead
            os.system ("python setup.py sdist")         # build sdist
            os.system ("cp 'dist/%s' '%s/%s'" % \
                    (sdist_name, mod_root, sdist_name)) # copy into tree
            os.system ("mv VERSION.bak VERSION")        # restore version

        print 'creating %s/SDIST' % path
        with open (path + "/SDIST", "w") as f : f.write (sdist_name + "\n")

        return version, version_detail, sdist_name

    except Exception as e :
        raise RuntimeError ('Could not extract/set version: %s' % e)


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_detail, sdist_name = get_version (mod_root)


# ------------------------------------------------------------------------------
# check python version. we need > 2.6, <3.x
if  sys.hexversion < 0x02060000 or sys.hexversion >= 0x03000000:
    raise RuntimeError('%s requires Python 2.x (2.6 or higher)' % name)


# ------------------------------------------------------------------------------
class our_test(Command):
    user_options = []
    def initialize_options (self) : pass
    def finalize_options   (self) : pass
    def run (self) :
        testdir = "%s/tests/" % os.path.dirname(os.path.realpath(__file__))
        retval  = sp.call([sys.executable,
                          '%s/run_tests.py'               % testdir,
                          '%s/configs/default.cfg'        % testdir])
        raise SystemExit(retval)


# ------------------------------------------------------------------------------
#
def read(*rnames):
    try :
        return open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    except Exception :
        return ''


# -------------------------------------------------------------------------------
c_ext = Extension("radical/synapse/atoms/_atoms" ,
                  sources = glob.glob('src/radical/synapse/atoms/*.c'))
setup_args = {
    'name'               : name,
    'version'            : version,
    'description'        : 'SYNthetic APplicationS Emulator -- A RADICAL Project '
                           '(http://radical.rutgers.edu/)',
    'long_description'   : (read('README.md') + '\n\n' + read('CHANGES.md')),
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'https://www.github.com/radical-cybertools/radical.utils/',
    'license'            : "LGPLv3+",
    'keywords'           : "radical emulate workload",
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    
    'ext_modules'        : [c_ext],
    'namespace_packages' : ['radical'],
    'packages'           : find_packages('src'),
    'package_dir'        : {'': 'src'},
    'scripts'            : ['bin/radical-synapse-version',
                            'bin/radical-synapse-profile',
                            'bin/radical-synapse-stats',
                            'bin/radical-synapse-stats.plot',
                            'bin/radical-synapse-execute',
                            'bin/radical-synapse-emulate',
                            'bin/radical-synapse-fixdocs',
                            'bin/radical-synapse-mandelbrot-dummy.py',
                            'bin/radical-synapse-mandelbrot-master.py',
                            'bin/radical-synapse-mandelbrot-worker.py',
                            'bin/radical-synapse-mandelbrot-profile.py',
                            'bin/radical-synapse-mandelbrot-emulate.py',
                            'bin/radical-synapse-iotrace.sh'
                           ],
    'package_data'       : {'': ['*.sh', '*.json', '*.gz', 'VERSION', 'SDIST', sdist_name, '*.c', '*.h']},
    'cmdclass'           : {
        'test'           : our_test,
    },
    'install_requires'   : [
        'pymongo', 
        'radical.utils', 
        'psutil'
    ],
    'extras_require'     : {
    },
    'tests_require'      : [],
    'test_suite'         : 'radical.synapse.tests',
    'zip_safe'           : False,
#   'build_sphinx'       : {
#       'source-dir'     : 'docs/',
#       'build-dir'      : 'docs/build',
#       'all_files'      : 1,
#   },
#   'upload_sphinx'      : {
#       'upload-dir'     : 'docs/build/html',
#   }
}

# ------------------------------------------------------------------------------

setup (**setup_args)

# ------------------------------------------------------------------------------

